/*
 * @(#)$Id: x11.c,v 1.5 1996/04/21 02:32:06 twitham Exp $
 *
 * Copyright (C) 1996 Tim Witham <twitham@pcocd2.intel.com>
 *
 * (see the files README and COPYING for more details)
 *
 * This file implements the X11 specific pieces of the display
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "oscope.h"		/* program defaults */
#include "x11.h"
#include "display.h"
#include "func.h"
#include "file.h"

Widget draw_widget;		/* scope drawing area */
Widget file[4];			/* file menu */
Widget plot[5];			/* plot menu */
Widget grat[6];			/* graticule menu */
Widget colormenu[17];		/* color menu */
Widget xwidg[5];		/* extra horizontal widgets */
Widget mwidg[56];		/* memory / math widgets */
Widget cwidg[CHANNELS];		/* channel button widgets */
Widget ywidg[15];		/* vertical widgets */
Widget **math;			/* math menu */
int **intarray;			/* indexes of math functions */
int XX[] = {640,800,1024,1280};
int XY[] = {480,600, 768,1024};
XFont font;
int color[16];
char *alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
char *colors[] = {		/* X colors similar to 16 console colors */
  "black",			/* 0 */
  "blue",
  "green",			/* 2 */
  "cyan",
  "red",			/* 4 */
  "magenta",
  "orange",			/* 6 */
  "gray66",
  "gray33",			/* 8 */
  "blue4",
  "green4",			/* 10 */
  "cyan4",
  "red4",			/* 12 */
  "magenta4",
  "yellow",			/* 14 */
  "white"
};

/* a libsx text writer similar to libvgamisc's vga_write */
int
VGA_WRITE(char *s, short x, short y, XFont f, short fg, short bg, char p)
{
  SetColor(fg);
  if (p == ALIGN_CENTER)
    x -= (TextWidth(f, s) / 2);
  else if (p == ALIGN_RIGHT)
    x -= (TextWidth(f, s));
  DrawText(s, x, y + FontHeight(f));
  return(1);
}

/* callback to redisplay the drawing area; snap to a graticule division */
void
redisplay(Widget w, int new_width, int new_height, void *data) {
  static int i = 2, h, v;

  h = h_points;
  v = v_points;
  h_points = new_width > 640
    ? new_width - (new_width - 200) % 44
    : 640;
  v_points = new_height > 480
    ? new_height - (new_height - 160) % 64
    : 480;
  offset = v_points / 2;
  if (h != h_points || v != v_points)
    ClearDrawArea();
  /* redo text later after widget redraws have settled down */
  AddTimeOut(200, draw_text, &i);
}

/* callback for keypress events on the scope drawing area */
void
keys(Widget w, char *input, int up_or_down, void *data)
{
  if (!up_or_down)		/* 0 press, 1 release */
    return;
  if (input[1] == '\0')		/* single character to handle */
    handle_key(input[0]);
}

/* menu option callbacks */
void
loadfile(Widget w, void *data)
{
  char *fname;

  fname = GetFile(NULL);
  if (fname != NULL)
    readfile(fname);
}

void
savefile(Widget w, void *data)
{
  char *fname;

  fname = GetFile(NULL);
  if (fname != NULL)
      writefile(fname);
}

void
plotmode(Widget w, void *data)
{
  char *c = (char *)data;

  scope.mode = *c - '0';
  clear();
}

void
graticule(Widget w, void *data)
{
  char *c = (char *)data;
  int i;

  i = *c - '0';
  if (i < 2)
    scope.behind = i;
  else
    scope.grat = i - 2;
  clear();
}

void
mathselect(Widget w, void *data)
{
  int *c = (int *)data;

  if (scope.select > 1) {
    ch[scope.select].func = *c + 4;
    clear();
  }
}

/* close the window */
void
dismiss(Widget w, void *data)
{
  CloseWindow();
}

/* slurp the manual page into a window */
void
help(Widget w, void *data)
{
  Widget x[2];
  char c, prev = '\0', pprev = '\0', *tmp, *s = NULL;
  FILE *p;
  int i = 0;

  if ((p = popen(HELPCOMMAND, "r")) != NULL) {
    while ((c = fgetc(p)) != EOF) {
      if (c == '\b')
	fgetc(p);
      else if (!(c == '\n' && prev == '\n' && pprev == '\n')) {
	tmp = realloc(s, i + 1);
	s = tmp;
	s[i] = c;
	i++;
      }
      pprev = prev;
      prev = c;
    }
    tmp = realloc(s, i + 1);
    s = tmp;
    s[i] = '\0';
    pclose(p);
    MakeWindow("Help", SAME_DISPLAY, NONEXCLUSIVE_WINDOW);
    x[0] = MakeButton("Dismiss", dismiss, NULL);
    x[1] = MakeTextWidget(s, FALSE, FALSE, 500, 500);
    free(s);
    SetWidgetPos(x[1], PLACE_UNDER, x[0], NO_CARE, NULL);
    ShowDisplay();
  }
}

/* simple button callback that just hits the given key */
void
hit_key(Widget w, void *data)
{
  char *c = (char *)data;

  handle_key(*c);
}

/* die on malloc error */
void
nomalloc(int line)
{
  sprintf(error, "%s: out of memory at x11.c line %d!", progname, line);
  perror(error);
  exit(1);
}

void
cleanup_x11()
{
  int i;

  for (i = 0 ; i < (funccount > 19 ? funccount - 4 : 16) ; i++) {
    if (i < funccount - 4)
      free(math[i]);
    free(intarray[i]);
  }
  free(math);
  free(intarray);
  FreeFont(font);
}

/* set current state colors, labels, and check marks on widgets */
void
fix_widgets()
{
  int i;
  char *tilt[] = {"-", "/", "\\"};

  for (i = 0 ; i < 4 ; i++) {
    SetMenuItemChecked(plot[i + 1], scope.mode == i);
  }
  SetMenuItemChecked(grat[1], !scope.behind);
  SetMenuItemChecked(grat[2], scope.behind);
  for (i = 0 ; i < 3 ; i++) {
    SetMenuItemChecked(grat[i + 3], scope.grat == i);
  }
  for (i = 0 ; i < 16 ; i++) {
    SetMenuItemChecked(colormenu[i + 1], scope.color == i);
  }
  for (i = 0 ; i < funccount - 4 ; i++) {
    SetMenuItemChecked(*math[i], ch[scope.select].func == i + 4);
  }

  SetBgColor(mwidg[0], ch[scope.select].color);
  SetBgColor(mwidg[27], ch[scope.select].color);
  SetBgColor(mwidg[28], ch[scope.select].color);
  SetBgColor(mwidg[55], ch[scope.select].color);
  SetWidgetState(mwidg[0], actual >= 44000);
  SetWidgetState(mwidg[27], scope.select > 1);
  SetWidgetState(mwidg[28], scope.select > 1);
  SetWidgetState(mwidg[55], scope.select > 1);
  for (i = 0 ; i < 26 ; i++) {
    if (mem[i] != NULL)
      SetBgColor(mwidg[i + 29], memcolor[i]);
  }

  SetLabel(xwidg[4], scope.run ? "Stop" : "Run");

  SetBgColor(ywidg[2], ch[scope.trigch].color);
  SetBgColor(ywidg[3], ch[scope.trigch].color);
  SetBgColor(ywidg[4], ch[scope.trigch].color);
  SetBgColor(ywidg[5], ch[scope.trigch].color);
  SetBgColor(ywidg[7], ch[scope.select].color);
  SetBgColor(ywidg[8], ch[scope.select].color);
  SetBgColor(ywidg[11], ch[scope.select].color);
  SetBgColor(ywidg[12], ch[scope.select].color);
  SetBgColor(ywidg[14], ch[scope.select].color);

  SetLabel(ywidg[3], scope.trigch ? "2" : "1");
  SetLabel(ywidg[4], tilt[scope.trige]);
  SetLabel(ywidg[14], ch[scope.select].show ? "Hide" : "Show");

  for (i = 0 ; i < CHANNELS ; i++) {
    SetFgColor(cwidg[i], ch[i].show ? color[0] : ch[i].color);
    SetBgColor(cwidg[i], ch[i].show ? ch[i].color : color[0]);
  }
}

void
setcolor(Widget w, void *data)
{
  int *c = (int *)data;

  scope.color = *c;
  draw_text(1);
}

/* initialize all the widgets, called by init_screen in display.c */
void
init_widgets()
{
  static char *s[] = {" 1  ", " 2  ", " 3  ", " 4  ",
		      " 5  ", " 6  ", " 7  ", " 8  "};
  int i, j;

  /* top row of widgets */
  file[0] = MakeMenu(" File ");
  file[1] = MakeMenuItem(file[0], "Load...", loadfile, NULL);
  file[2] = MakeMenuItem(file[0], "Save...", savefile, NULL);
  file[3] = MakeMenuItem(file[0], "Quit", hit_key, "\e");

  plot[0] = MakeMenu(" Plot Mode ");
  plot[1] = MakeMenuItem(plot[0], "Point", plotmode, "0");
  plot[2] = MakeMenuItem(plot[0], "Point Accumulate", plotmode, "1");
  plot[3] = MakeMenuItem(plot[0], "Line", plotmode, "2");
  plot[4] = MakeMenuItem(plot[0], "Line Accumulate", plotmode, "3");

  colormenu[0] = MakeMenu("Color");

  grat[0] = MakeMenu(" Graticule ");
  grat[1] = MakeMenuItem(grat[0], "In Front", graticule, "0");
  grat[2] = MakeMenuItem(grat[0], "Behind", graticule, "1");
  grat[3] = MakeMenuItem(grat[0], "None", graticule, "2");
  grat[4] = MakeMenuItem(grat[0], "Minor Divisions", graticule, "3");
  grat[5] = MakeMenuItem(grat[0], "Minor & Major", graticule, "4");

  xwidg[0] = MakeButton(" Refresh ", hit_key, "\n");
  xwidg[1] = MakeButton(" < ", hit_key, "9");
  xwidg[2] = MakeLabel("Sec/Div");
  xwidg[3] = MakeButton(" > ", hit_key, "0");
  xwidg[4] = MakeButton("Stop", hit_key, " ");

  SetWidgetPos(plot[0], PLACE_RIGHT, file[0], NO_CARE, NULL);
  SetWidgetPos(grat[0], PLACE_RIGHT, plot[0], NO_CARE, NULL);
  SetWidgetPos(colormenu[0], PLACE_RIGHT, grat[0], NO_CARE, NULL);
  SetWidgetPos(xwidg[0], PLACE_RIGHT, colormenu[0], NO_CARE, NULL);
  SetWidgetPos(xwidg[1], PLACE_RIGHT, xwidg[0], NO_CARE, NULL);
  SetWidgetPos(xwidg[2], PLACE_RIGHT, xwidg[1], NO_CARE, NULL);
  SetWidgetPos(xwidg[3], PLACE_RIGHT, xwidg[2], NO_CARE, NULL);
  SetWidgetPos(xwidg[4], PLACE_RIGHT, xwidg[3], NO_CARE, NULL);

  /* the drawing area for the scope */
  h_points = XX[scope.size];
  v_points = XY[scope.size];
  draw_widget = MakeDrawArea(h_points, v_points, redisplay, NULL);
  SetKeypressCB(draw_widget, keys);
  SetWidgetPos(draw_widget, PLACE_UNDER, file[0], NO_CARE, NULL);

  /* right column of widgets */
  ywidg[0] = MakeButton("Help", help, NULL);

  ywidg[1] = MakeLabel("Trig");
  ywidg[2] = MakeButton(" /\\ ", hit_key, "=");
  ywidg[3] = MakeButton("<", hit_key, "_");
  ywidg[4] = MakeButton(">", hit_key, "+");
  ywidg[5] = MakeButton(" \\/ ", hit_key, "-");

  ywidg[6] = MakeLabel("Pos.");
  ywidg[7] = MakeButton(" /\\ ", hit_key, "}");
  ywidg[8] = MakeButton(" \\/ ", hit_key, "{");

  ywidg[9] = MakeLabel("Chan");

  SetWidgetPos(ywidg[0],  PLACE_RIGHT, draw_widget, NO_CARE, NULL);
  SetWidgetPos(ywidg[1],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[0]);
  SetWidgetPos(ywidg[2],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[1]);
  SetWidgetPos(ywidg[3],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[2]);
  SetWidgetPos(ywidg[4],  PLACE_RIGHT, ywidg[3], PLACE_UNDER, ywidg[2]);
  SetWidgetPos(ywidg[5],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[4]);
  SetWidgetPos(ywidg[6],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[5]);
  SetWidgetPos(ywidg[7],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[6]);
  SetWidgetPos(ywidg[8],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[7]);
  SetWidgetPos(ywidg[9],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[8]);
  
  for (i = 0 ; i < CHANNELS ; i++) {
    cwidg[i] = MakeButton(s[i], hit_key, &s[i][1]);
    SetWidgetPos(cwidg[i],  PLACE_RIGHT, draw_widget,
		 PLACE_UNDER, i ? cwidg[i - 1] : ywidg[9]);
  }

  ywidg[10] = MakeLabel("Scal");
  ywidg[11] = MakeButton(" /\\ ", hit_key, "]");
  ywidg[12] = MakeButton(" \\/ ", hit_key, "[");
  ywidg[13] = MakeLabel("");
  ywidg[14] = MakeButton("Hide", hit_key, "\t");

  SetWidgetPos(ywidg[10],  PLACE_RIGHT, draw_widget,
	       PLACE_UNDER, cwidg[CHANNELS - 1]);
  SetWidgetPos(ywidg[11],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[10]);
  SetWidgetPos(ywidg[12],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[11]);
  SetWidgetPos(ywidg[13],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[12]);
  SetWidgetPos(ywidg[14],  PLACE_RIGHT, draw_widget, PLACE_UNDER, ywidg[13]);

  /* bottom rows of widgets */
  mwidg[0]  = MakeLabel(" Store ");
  mwidg[28] = MakeLabel(" Recall");
  SetWidgetPos(mwidg[0],  PLACE_UNDER, draw_widget, NO_CARE, NULL);
  SetWidgetPos(mwidg[28],  PLACE_UNDER, mwidg[0], NO_CARE, NULL);
  for (i = 0 ; i < 26 ; i++) {
    sprintf(error, "%c", i + 'A');
    mwidg[i + 1] = MakeButton(error, hit_key, &alphabet[i]);
    SetWidgetPos(mwidg[i + 1], PLACE_UNDER, draw_widget,
		 PLACE_RIGHT, mwidg[i]);
    sprintf(error, "%c", i + 'a');
    mwidg[i + 29] = MakeButton(error, hit_key, &alphabet[i + 26]);
    SetWidgetPos(mwidg[i + 29], PLACE_UNDER, mwidg[0],
		 PLACE_RIGHT, mwidg[i + 28]);
  }
  mwidg[27] = MakeMenu(  "Math     ");
  mwidg[55] = MakeButton("Extern...", hit_key, "$");
  SetWidgetPos(mwidg[27],  PLACE_UNDER, draw_widget, PLACE_RIGHT, mwidg[26]);
  SetWidgetPos(mwidg[55],  PLACE_UNDER, mwidg[27], PLACE_RIGHT, mwidg[54]);

  j = funccount - 4;
  if ((math = malloc(sizeof(Widget *) * j)) == NULL)
    nomalloc(__LINE__);
  if ((intarray = malloc(sizeof(int *) * (j > 16 ? j : 16))) == NULL)
    nomalloc(__LINE__);
  for (i = 0 ; i < (j > 16 ? j : 16) ; i++) {
    if ((intarray[i] = malloc(sizeof(int))) == NULL)
      nomalloc(__LINE__);
    *intarray[i] = i;
    if (i < j) {
      if ((math[i] = malloc(sizeof(Widget))) == NULL)
	nomalloc(__LINE__);
      *math[i] = MakeMenuItem(mwidg[27], funcnames[4 + i],
			      mathselect, intarray[i]);
    }
  }

  ShowDisplay();

  for (i = 0 ; i < 16 ; i++) {
    color[i] = GetNamedColor(colors[i]);
    colormenu[i + 1] = MakeMenuItem(colormenu[0], colors[i],
				    setcolor, intarray[i]);
  }

  for (i = 0 ; i < 26 ; i++) {
    SetBgColor(mwidg[i + 29], color[0]);
  }

  SetBgColor(draw_widget, color[0]);
  font = GetFont(fontname);
  SetWidgetFont(draw_widget, font);
  ClearDrawArea();
}
